import warnings

import numpy as np

__all__ = ["_reliability"]


def _reliability(o, f, bin_edges):
    """Return the reliability and number of samples per bin"""
    # I couldn't get dask='parallelized' working in this case
    # so dealing with dask arrays explicitly
    # if true, imports dask.array later
    is_dask_array = not isinstance(o, np.ndarray) or not isinstance(f, np.ndarray)

    if is_dask_array:
        r = []
        N = []
    else:
        r = np.zeros((*o.shape[:-1], len(bin_edges) - 1), dtype=float)
        N = np.zeros_like(r)

    for i in range(len(bin_edges) - 1):
        # Follow xhistogram behaviour: all bins are half-open,
        # except for the right-most bin
        # which adds an epsilon to the right edge
        # see https://github.com/xgcm/xhistogram/issues/18
        if i == (len(bin_edges) - 2):
            f_in_bin = (f >= bin_edges[i]) & (f < (bin_edges[i + 1] + 1e-8))
        else:
            f_in_bin = (f >= bin_edges[i]) & (f < bin_edges[i + 1])
        o_f_in_bin = o & f_in_bin
        N_f_in_bin = f_in_bin.sum(axis=-1)
        N_o_f_in_bin = o_f_in_bin.sum(axis=-1)
        if is_dask_array:
            r.append(N_o_f_in_bin / N_f_in_bin)
            N.append(N_f_in_bin)
        else:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", RuntimeWarning)
                r[..., i] = N_o_f_in_bin / N_f_in_bin
                N[..., i] = N_f_in_bin

    if is_dask_array:
        import dask.array as da

        return (
            da.stack(r, axis=-1).rechunk({-1: -1}),
            da.stack(N, axis=-1).rechunk({-1: -1}),
        )
    else:
        return r, N



from __future__ import annotations

from typing import List, Union

import xarray as xr

XArray = Union[xr.Dataset, xr.DataArray]
# XArray = xr.Dataset | xr.DataArray
# raises error during build:
# TypeError: unsupported operand type(s) for |: 'ABCMeta' and 'type'
Dim = Union[List[str], str]
# Dim = List[str] | str



import pytest
from dask import is_dask_collection

import xskillscore as xs
from xskillscore import multipletests


@pytest.fixture
def r_p(a, b):
    return xs.pearson_r_p_value(a, b, "time")


@pytest.mark.parametrize("chunk", [True, False])
@pytest.mark.parametrize("input", ["Dataset", "multidim Dataset", "DataArray"])
def test_multipletests_inputs(r_p, input, chunk):
    """Test multipletests with xr inputs and chunked."""
    method = "fdr_bh"
    alpha = 0.05
    if "Dataset" in input:
        name = "var"
        r_p = r_p.to_dataset(name=name)
        if input == "multidim Dataset":
            r_p["var2"] = r_p["var"] * 2
    if chunk:
        r_p = r_p.chunk()
    ret = multipletests(r_p, method=method, alpha=alpha, return_results="all_as_result_dim")
    # check dask collection preserved
    assert is_dask_collection(ret) if chunk else not is_dask_collection(ret)
    # check coords added
    assert ret.coords["multipletests_method"] == method
    assert ret.coords["multipletests_alpha"] == alpha


def test_multipletests_alpha(r_p):
    """Test that larger alpha leads to more rejected in multipletests."""
    method = "fdr_bh"
    reject = multipletests(r_p, alpha=0.05, method=method).sel(result="reject")
    reject_larger_alpha = multipletests(r_p, alpha=0.5, method=method).sel(result="reject")
    # check more reject
    assert reject_larger_alpha.sum() > reject.sum()


def test_multipletests_method(r_p):
    """Test that multipletests is sensitive to method."""
    ret_fdr_bh = multipletests(r_p, method="fdr_bh")
    ret_hs = multipletests(r_p, method="hs")
    assert not ret_hs.equals(ret_fdr_bh)


@pytest.mark.parametrize(
    "method",
    [
        "bonferroni",
        "sidak",
        "holm-sidak",
        "hs",
        "holm",
        "simes-hochberg",
        "hommel",
        "fdr_bh",
        "fdr_by",
        "fdr_tsbh",
        "fdr_tsbky",
    ],
)
def test_multipletests_all_methods(r_p, method):
    """Test that multipletests accepts all methods."""
    assert multipletests(r_p, method=method).any()


def test_multipletests_return_results(r_p):
    """Test that multipletests is sensitive to return_results."""
    for return_results in ["pvals_corrected", "all_as_tuple", "all_as_result_dim"]:
        ret = multipletests(r_p, method="fdr_bh", return_results=return_results)
        if return_results == "pvals_corrected":
            assert isinstance(ret, type(r_p))
            assert "result" not in ret.dims
            assert ret.coords["result"] == "pvals_corrected"
        elif return_results == "all_as_tuple":
            assert isinstance(ret, tuple)
            assert all(isinstance(r, type(r_p)) for r in ret)
            assert ret[1].coords["result"] == "pvals_corrected"
        elif return_results == "all_as_result_dim":
            assert isinstance(ret, type(r_p))
            assert "result" in ret.dims
            assert "pvals_corrected" in ret.result


@pytest.mark.parametrize(
    "kwargs",
    [
        dict(method=None),
        dict(method="doesnotexist"),
        dict(method="fdr_bh", alpha=".5"),
        dict(method="fdr_bh", alpha=[]),
        dict(method="fdr_bh", alpha=True),
        dict(method="fdr_bh", return_results=None),
        dict(method="fdr_bh", return_results=True),
    ],
)
def test_multipletests_fails(r_p, kwargs):
    with pytest.raises(ValueError):
        multipletests(r_p, **kwargs)



# Write the benchmarking functions here.
# See "Writing benchmarks" in the asv docs for more information.


import bottleneck as bn
import numpy as np
import xarray as xr

from xskillscore import (
    mae as xs_mae,
)
from xskillscore import (
    mse as xs_mse,
)
from xskillscore import (
    pearson_r as xs_pearson_r,
)
from xskillscore import (
    rmse as xs_rmse,
)
from xskillscore import (
    spearman_r as xs_spearman_r,
)

from . import Generate, parameterized, requires_dask

# These metrics in xskillscore.xr.deterministic, entirely written in xarray functions,
# are identical to the well documented metrics in xskillscore.core.deterministic, which
# are based on numpy functions applied to xarray objects by xarray.apply_ufunc. As the
# xr metrics are only faster for small data, their use is not encouraged, as the
# numpy-based metrics are 20-40% faster on large data."""


def xr_mse(a, b, dim=None, skipna=True, weights=None):
    res = (a - b) ** 2
    if weights is not None:
        res = res.weighted(weights)
    res = res.mean(dim=dim, skipna=skipna)
    return res


def xr_mae(a, b, dim=None, skipna=True, weights=None):
    res = np.abs(a - b)
    if weights is not None:
        res = res.weighted(weights)
    res = res.mean(dim=dim, skipna=skipna)
    return res


def xr_me(a, b, dim=None, skipna=True, weights=None):
    res = a - b
    if weights is not None:
        res = res.weighted(weights)
    res = res.mean(dim=dim, skipna=skipna)
    return res


def xr_rmse(a, b, dim=None, skipna=True, weights=None):
    res = (a - b) ** 2
    if weights is not None:
        res = res.weighted(weights)
    res = res.mean(dim=dim, skipna=skipna)
    res = np.sqrt(res)
    return res


def xr_pearson_r(a, b, dim=None, **kwargs):
    return xr.corr(a, b, dim)


def _rankdata(o, dim):
    if isinstance(dim, str):
        dim = [dim]
    elif dim is None:
        dim = list(o.dims)
    if len(dim) == 1:
        return xr.apply_ufunc(
            bn.nanrankdata,
            o,
            input_core_dims=[[]],
            kwargs={"axis": o.get_axis_num(dim[0])},
            dask="allowed",
        )
    elif len(dim) > 1:
        # stack rank unstack
        return xr.apply_ufunc(
            bn.nanrankdata,
            o.stack(ndim=dim),
            input_core_dims=[[]],
            kwargs={"axis": -1},
            dask="allowed",
        ).unstack("ndim")


def xr_spearman_r(a, b, dim=None, **kwargs):
    return xr.corr(_rankdata(a, dim), _rankdata(b, dim), dim)


METRICS = [
    xs_mse,
    xr_mse,
    xs_rmse,
    xr_rmse,
    xs_mae,
    xr_mae,
    xs_pearson_r,
    xr_pearson_r,
    xs_spearman_r,
    xr_spearman_r,
]

DIMS = ["member", ["lon", "lat"]]

large_lon_lat = 2000
large_lon_lat_chunksize = large_lon_lat // 4
ntime = 24


class Compute_small(Generate):
    """
    A benchmark xskillscore.metric for small xr.DataArrays"""

    def setup(self, *args, **kwargs):
        self.make_ds(ntime, 1, 1)  # no grid

    @parameterized(["metric", "dim"], (METRICS, DIMS))
    def time_xskillscore_metric(self, metric, dim):
        """Take time for xskillscore.metric."""
        metric(self.ds["tos"], self.ds["sos"], dim=dim).compute()

    @parameterized(["metric", "dim"], (METRICS, DIMS))
    def peakmem_xskillscore_metric(self, metric, dim):
        """Take memory peak for xskillscore.metric."""
        metric(self.ds["tos"], self.ds["sos"], dim=dim).compute()


class Compute_large(Compute_small):
    """
    A benchmark xskillscore.metric for large xr.DataArrays"""

    def setup(self, *args, **kwargs):
        self.make_ds(ntime, large_lon_lat, large_lon_lat)


class Compute_large_dask(Compute_small):
    """
    A benchmark xskillscore.metric for large xr.DataArrays with dask."""

    def setup(self, *args, **kwargs):
        requires_dask()
        self.make_ds(ntime, large_lon_lat, large_lon_lat, chunks={"lon": large_lon_lat_chunksize})



from __future__ import annotations

import warnings
from typing import Literal, Mapping, Optional, Tuple, Union

import xarray as xr
from statsmodels.stats.multitest import multipletests as statsmodels_multipletests

from .types import XArray


def multipletests(
    p: XArray,
    alpha: float = 0.05,
    method: Optional[
        Literal[
            "bonferroni",
            "sidak",
            "holm-sidak",
            "hs",
            "holm",
            "simes-hochberg",
            "hommel",
            "fdr_bh",
            "fdr_by",
            "fdr_tsbh",
            "fdr_tsbky",
        ]
    ] = None,
    keep_attrs=True,
    return_results: Literal[
        "pvals_corrected", "all_as_result_dim", "all_as_tuple"
    ] = "all_as_result_dim",
    **multipletests_kwargs: Mapping,
) -> Union[XArray, Tuple[XArray, ...]]:
    """Apply statsmodels.stats.multitest.multipletests for controlling the false
    discovery rate for multiple hypothesis tests for multidimensional
    xr.DataArray and xr.Datasets.

    Parameters
    ----------
    p : xarray.Dataset or xarray.DataArray
        uncorrected p-values.
    alpha : float, optional
        family-wise error rate (FWER). Defaults to 0.05.
    method : str
        Method used for testing and adjustment of pvalues. Can be
        either the full name or initial letters. Available methods are:
        - "bonferroni" : one-step correction
        - "sidak" : one-step correction
        - "holm-sidak" : step down method using Sidak adjustments (alias "hs")
        - "holm" : step-down method using Bonferroni adjustments
        - "simes-hochberg" : step-up method (independent)
        - "hommel" : closed method based on Simes tests (non-negative)
        - "fdr_bh" : Benjamini/Hochberg (non-negative)
        - "fdr_by" : Benjamini/Yekutieli (negative)
        - "fdr_tsbh" : two stage fdr correction (non-negative)
        - "fdr_tsbky" : two stage fdr correction (non-negative)
    keep_attrs : bool
        If True, the attributes (attrs) will be copied
        from the first input to the new one.
        If False (default), the new object will
        be returned without attributes.
    return_results : str
        Specify how return is structed:

            - 'pvals_corrected': return only the corrected p values

            - 'all_as_tuple': return (reject, pvals_corrected, alphacSidak,
              alphacBonf) as tuple

            - 'all_as_result_dim': return (reject, pvals_corrected, alphacSidak,
              alphacBonf) concatenated into new ``result`` dimension. (Default)

    **multipletests_kwargs : dict, optional
        is_sorted, returnsorted, see statsmodels.stats.multitest.multitest

    Returns
    -------
    reject : xarray.Dataset or xarray.DataArray
        true for hypothesis that can be rejected for given alpha
    pvals_corrected : xarray.Dataset or xarray.DataArray
        p-values corrected for multiple tests
    alphacSidak : xarray.Dataset or xarray.DataArray
        corrected alpha for Sidak method
    alphacBonf : xarray.Dataset or xarray.DataArray
        corrected alpha for Bonferroni method

    References
    ----------
    * Wilks, D. S. (2016). “The Stippling Shows Statistically Significant Grid Points”:
      How Research Results are Routinely Overstated and Overinterpreted, and What to Do
      about It. Bulletin of the American Meteorological Society, 97(12), 2263–2273.
      https://www.doi.org/10/f9mvth
    * Benjamini, Y., & Hochberg, Y. (1994). Controlling the False Discovery Rate:
      A Practical and Powerful Approach to Multiple Testing.
      Journal of the Royal Statistical Society: Series B (Methodological), 57(1), 13.
      https://www.doi.org/10.1111/j.2517-6161.1995.tb02031.x

    Examples
    --------
    >>> p = xr.DataArray(
    ...     np.random.normal(size=(3, 3)),
    ...     coords=[("x", np.arange(3)), ("y", np.arange(3))],
    ... )
    >>> result = xs.multipletests(p, alpha=0.1, method="fdr_bh", return_results="all_as_result_dim")
    >>> result
    <xarray.DataArray (result: 4, x: 3, y: 3)> Size: 288B
    array([[[ 0.        ,  1.        ,  0.        ],
            [ 0.        ,  1.        ,  1.        ],
            [ 0.        ,  0.        ,  1.        ]],
    <BLANKLINE>
           [[ 0.49671415, -0.1382643 ,  0.64768854],
            [ 1.        , -0.23415337, -0.23413696],
            [ 1.        ,  0.76743473, -0.46947439]],
    <BLANKLINE>
           [[ 0.1       ,  0.1       ,  0.1       ],
            [ 0.1       ,  0.1       ,  0.1       ],
            [ 0.1       ,  0.1       ,  0.1       ]],
    <BLANKLINE>
           [[ 0.1       ,  0.1       ,  0.1       ],
            [ 0.1       ,  0.1       ,  0.1       ],
            [ 0.1       ,  0.1       ,  0.1       ]]])
    Coordinates:
      * result                (result) <U15 240B 'reject' ... 'alphacBonf'
      * x                     (x) int64 24B 0 1 2
      * y                     (y) int64 24B 0 1 2
        multipletests_method  <U6 24B 'fdr_bh'
        multipletests_alpha   float64 8B 0.1
    """
    MULTIPLE_TESTS = [
        "bonferroni",
        "sidak",
        "holm-sidak",
        "hs",
        "holm",
        "simes-hochberg",
        "hommel",
        "fdr_bh",
        "fdr_by",
        "fdr_tsbh",
        "fdr_tsbky",
    ]
    msg = "Alpha must be float between 0.0 and 1.0."
    if not isinstance(alpha, float):
        raise ValueError(msg)
    elif alpha <= 0.0 or alpha >= 1.0:
        raise ValueError(msg)

    if method is None:
        raise ValueError(
            f"Please indicate a method using the 'method=...' keyword. Select from {MULTIPLE_TESTS}"
        )
    elif method not in MULTIPLE_TESTS:
        raise ValueError(f"Your method '{method}' is not in the accepted methods: {MULTIPLE_TESTS}")

    allowed_return_results = ["all_as_tuple", "pvals_corrected", "all_as_result_dim"]
    if return_results not in allowed_return_results:
        raise ValueError(
            f"Expected `return_results` from {allowed_return_results}, found {return_results}"
        )

    # Suppress NumPy scalar conversion deprecation warning from internal numpy operations
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message="Conversion of an array with ndim > 0 to a scalar is deprecated",
            category=DeprecationWarning,
        )
        ret = xr.apply_ufunc(
            statsmodels_multipletests,
            p.stack(s=p.dims),
            input_core_dims=[[]],
            vectorize=True,
            output_core_dims=[[]] * 4,
            output_dtypes=[bool, float, float, float],
            kwargs=dict(method=method, alpha=alpha, **multipletests_kwargs),
            dask="parallelized",
            keep_attrs=keep_attrs,
        )

    ret = tuple(r.unstack("s").transpose(*p.dims, ...) for r in ret)

    def _add_kwargs_as_coords(r: XArray):
        return r.assign_coords(multipletests_method=method, multipletests_alpha=alpha)

    ret = tuple(_add_kwargs_as_coords(r) for r in ret)

    returns = ["reject", "pvals_corrected", "alphacSidak", "alphacBonf"]

    if return_results == "all_as_result_dim":
        return xr.concat(ret, "result").assign_coords(result=returns)
    elif return_results == "all_as_tuple":
        for i, r in enumerate(ret):
            r.coords["result"] = returns[i]
        return ret
    else:
        return ret[1].assign_coords(result="pvals_corrected")
